from spacy.language import Language 
from spacy.tokens import Doc

_de2c = {
  "ep": "主语补足语",
  "svp": "后缀",
  "oa": "第四格补足语",
  "da": "第三格补足语",
  "pd": "表语补足语",
  "ng": "否定词",
  "op": "介词补足语",
  "cvc": "介词短语",
  "og": "第二格补足语",
  "cj": "并列句",
  "ph": "形式主语",
  "rc": "关系从句",
  "rs": "其他",
  "vo": "其他",
  "cc": "其他",
  "app": "其他",
  "dep": "其他",
  "dm": "其他",
  "par": "其他",
  "sbp": "其他",
  "mnr": "其他",
}

trans_tag = {
  "VAFIN": "限定-助动词",
  "VAINF": "不定式-助动词",
  "VAIMP": "命令式-助动词",
  "VAPP": "过去分词-助动词",

  "VVFIN": "限定-动词",
  "VVINF": "不定式-动词",
  "VVIMP": "命令式-动词",
  "VVPP": "过去分词-动词",
  "VVIZU": "带zu不定式-动词",

  "VMFIN": "限定-情态动词",
  "VMINF": "不定式-情态动词",
  "VMPP": "过去分词-情态动词"
}

adp_verbs = {
  "an": ["arbeiten", "beteiligen", "erkennen", "erkranken", "fehlt", "hindern", "leiden", "liegt", "mangelt", "sterben", "teilnehmen", "vorbeikommen", "zweifeln", "denken", "erinnern", "gewöhnen", "glauben", "grenzen", "halten", "schicken", "schreiben", "vermieten", "wenden"],
  "auf": ["achten", "beschränken", "antworten", "aufpassen", "beziehen", "einigen", "freuen", "hoffen", "hören", "konzentrieren", "reagieren", "schätzen", "schießen", "schimpfen", "spezialisieren", "stoßen", "verlassen", "verzichten", "vorbereiten", "warten", "beruhen", "bestehen"],
  "aus": ["aussteigen", "bestehen", "lernen", "machen", "stammen", "übersetzen", "zusammensetzen"],
  "bei": ["bedanken", "beschweren", "entschuldigen", "helfen", "lernen"],
  "durch": ["ersetzen"],
  "für": ["arbeiten", "bedanken", "brauchen", "danken", "einsetzen", "eignen", "eintreten", "engagieren", "entscheiden", "entschuldigen", "halten", "interessieren", "kämpfen", "rächen", "schämen", "sorgen", "stimmen", "vormerken", "werben"],
  "gegen": ["kämpfen", "protestieren", "spielen", "verstoßen", "verteidigen", "wehren", "wenden"],
  "in": ["irren", "bestehen", "einmischen", "einsteigen", "geraten", "verlieben", "verwandeln"],
  "mit": ["anfangen", "aufhören", "auskommen", "beeilen", "befassen", "beginnen", "beschäftigen", "handeln", "kämpfen", "machen", "rechnen", "schimpfen", "telefonieren", "treffen", "verabreden", "verbinden", "vergleichen", "verheiraten", "verloben", "versöhnen", "verstehen", "vertauschen", "vertragen", "verwechseln", "warten", "zögern", "zusammenstoßen"],
  "nach": ["erkundigen", "fragen", "riechen", "schmecken", "sehnen", "suchen", "umhören"],
  "über": ["ärgern", "aufregen", "sagen", "beklagen", "berichten", "beschweren", "diskutieren", "erschrecken", "freuen", "informieren", "klagen", "lachen", "nachdenken", "reden", "schimpfen", "siegen", "sprechen", "staunen", "streiten", "unterhalten", "verfügen", "wissen", "wundern"],
  "um": ["bemühen", "beneiden", "bewerben", "bitten", "geht", "handelt", "kämpfen", "kümmern", "sorgen", "spielen", "streiten", "trauern", "werben"],
  "unter": ["leiden"],
  "von": ["abhängen", "befreien", "berichten", "erholen", "ernähren", "erwarten", "erzählen", "fordern", "halten", "handeln", "hören", "leben", "lernen", "sprechen", "trennen", "träumen", "überzeugen", "unterscheiden", "verabschieden", "verlangen", "wissen"],
  "vor": ["erschrecken", "fliehen", "flüchten", "fürchten", "schützen", "warnen"],
  "zu": ["auffordern", "beitragen", "beglückwünschen", "benutzen", "bringen", "einladen", "entschließen", "ernennen", "führen", "gehören", "gratulieren", "kommt", "passen", "überreden", "raten", "verurteilen", "wählen", "werden", "zwingen"]
}

adp_verbs["ans"] = adp_verbs["am"] = adp_verbs["an"]

adp_verbs["im"] = adp_verbs["ins"] = adp_verbs["in"]

adp_verbs["zum"] = adp_verbs["ins"] = adp_verbs["zu"]

adp_verbs["aufs"] = adp_verbs["auf"]

excluded_cop_adps = ["durch", "außer", "nach", "seit", "von", "wegen", "statt", "trotz", "während", "ab", "für"]
cop_verbs = ["sein", "werden", "bleiben", "belieb"]
aux_verbs = ["sein", "haben", "werden", "habe"]
decoration_advs = ["da", "doch", "bloß", "halt", "mal", "eben", "ja", "eh", "denn", "auch", "bitte"]

def _mo_adp_explanation(doc, ele, structure):
  #roots = [doc[s["element"]] for s in structure if doc[s["element"]].dep_=="ROOT"]
  #assert(len(roots) == 1)
  #root = roots[0]
  #oc = [s for s in structure if doc[s["element"]].dep_=="oc"]
  #deps = [doc[s["element"]].dep_ for s in structure]

  if ele.head.lemma_ in cop_verbs and ele.text.lower() not in excluded_cop_adps: # and all([d not in ["oc"] for d in deps])
    return "表语补足语"
  elif ele.text not in adp_verbs.keys():
    return "其他"
  elif ele.head.lemma_ in adp_verbs[ele.text]:
    return "介词补足语"
  else:
    return "其他"

  #elif len(oc)==0 and ele.head.lemma_ in adp_verbs[ele.text]:
  #  return "介词补足语"
  #elif len(oc)==1 and oc[0]["end"]-oc[0]["start"]==1 and doc[oc[0]["element"]].lemma_ in adp_verbs[ele.text]:
  #  return "介词补足语"

def _get_root_explanation(doc, r, structure):
  ele = doc[r["element"]]
  deps = [doc[s["element"]].dep_ for s in structure]
  explanation = None
  if ele.pos_ not in ["VERB", "AUX", "NOUN"]:
    explanation = None
  elif ele.pos_ == "AUX" and ele.tag_ in ["VMINF", "VMFIN", "VMPP"]:
    explanation = "情态动词"
  elif ele.lemma_ in aux_verbs and any([d in ["oc"] for d in deps]):
    explanation = "助动词"
  elif ele.lemma_ in cop_verbs:
    explanation = "系动词"
  elif ele.pos_ == "VERB": 
    explanation = "限定动词"
  elif ele.pos_ == "AUX":
    explanation = "助动词"
  else:
    explanation = ele.tag_
  return explanation

def _get_mo_explanation(doc, r, structure):
  ele = doc[r["element"]]
  explanation = None
  deps = [doc[s["element"]].dep_ for s in structure]
  if ele.tag_ in ["PWAT", "PWAV", "PWS"]:
    return "疑问词"
  elif ele.pos_ in ["ADP"]:
    return _mo_adp_explanation(doc, ele, structure)
  elif ele.head.lemma_ in ["heißen", "werden"] and ele.pos_ not in ["ADV"]:
    return "第一格补足语"
  elif ele.head.lemma_ in ["nennen", "kosten", "lehren", "schimpfen", "schelten"]:
    return "第四格补足语"
  else:
    return "其他"

def _get_oc_explanation(doc, r, structure):
  ele = doc[r["element"]]
  deps = [s.dep_ for s in ele.children]
  if ele.pos_ in ["AUX", "VERB"] and r["end"]-r["start"]==1:
    if ele.tag_ in trans_tag.keys():
      return trans_tag[ele.tag_]
    else:
      return ele.tag_
  elif ele.pos_ in ["PRON"] and r["end"]-r["start"]==1:
    return "第四格补足语"
  elif "pm" in deps and r["end"]-r["start"]==2:
    return "带zu不定式"
  elif "sb" in deps or "da" in deps or "oa" in deps or "ep" in deps:
    return "第四格补足语"
  elif ele.head.lemma_ in ["heiß", "heißen", "werden"]: # mo pos
    return "第一格补足语"
  else:
    return "其他"

def _get_sb_explanation(doc, t, structure):
  sbs = [s["element"] for s in structure if doc[s["element"]].dep_ == "sb"]
  if len(sbs) == 1:
    return "主语补足语"
  elif doc[t["element"]].pos_ == "PRON":
    return "主语补足语"
  else:
    return "表语补足语"

def _get_cd_explanation(doc, t, structure):
  ele = doc[t["element"]]
  if len(list(ele.children)) > 0:
    return "并列句"
  else:
    return "连词"

def _get_jucp_explanation(doc, t, structure):
  ele = doc[t["element"]]
  if ele.pos_ in ["SCONJ", "CCONJ"]:
    return "连词"
  else:
    return "其他"

def _get_refer(doc, t, structure):
  parent_id = doc[t["element"]].head.i
  parents = [p for p in structure if p["element"]==parent_id]
  assert(len(parents)==1)
  return parents[0]

def _analyze(doc, t, structure):
  dep = doc[t["element"]].dep_
  explanation = None
  if dep == "ROOT":
    explanation = _get_root_explanation(doc, t, structure)  
  elif dep == "mo":
    explanation = _get_mo_explanation(doc, t, structure)
  elif dep == "oc":
    explanation = _get_oc_explanation(doc, t, structure)
  elif dep == "re":
    t_refer = _get_refer(doc, t, structure)
    explanation = _analyze(doc, t_refer, structure)["explanation"]
    if explanation in ["疑问词"]:
      explanation = "其他"
  elif dep == "sb":
    explanation = _get_sb_explanation(doc, t, structure)
  elif dep == "cd":
    explanation = _get_cd_explanation(doc, t, structure)
  elif dep in ["ju", "cp"]:
    explanation = _get_jucp_explanation(doc, t, structure)
  elif dep in _de2c.keys():
    explanation = _de2c[dep]
  elif dep in ["punct"]:
    explanation = None
  else:
    explanation = dep
  t["explanation"] = explanation
  return t

def de_trs(doc, structure):
  structure_with_explanation = []
  for t in structure:
    _t = _analyze(doc, t, structure)
    structure_with_explanation.append(_t)
  return structure_with_explanation



from spacy.language import Language 
from spacy.tokens import Doc
from importlib import import_module

class Structure:

  def __init__(self, nlp: Language): 
    lang = nlp.meta["lang"]
    self.ext_name = "structure"
    self.pkg = 'structure_detective.{}.structure'.format(lang)
    self.nlp = nlp
    
    if not Doc.has_extension(self.ext_name):
      Doc.set_extension(self.ext_name, default={})

  def __call__(self, doc: Doc):
    pkg_module = import_module(self.pkg)    
    fn = getattr(pkg_module, 'get_tree')
    doc._.structure = fn(doc, self.nlp) 
    return doc

def trs(doc, structure: dict, lang: str):
  pkg = 'structure_detective.{}.trans_structure'.format(lang)
  pkg_module = import_module(pkg)    
  fn = getattr(pkg_module, '{}_trs'.format(lang))
  return fn(doc, structure)



from spacy.language import Language 
from spacy.tokens import Doc
from structure_detective.lib import get_subtree_except_xs, form_children_info

def _is_verb_like(e):
  return e.pos_ in ["VERB", "AUX"]

def _is_noun_like(e):
  return e.pos_ in ["NOUN", "ADJ", "ADV", "PRON", "DET"]

def _has_no_subject(e):
  children_deps = [x.dep_ for x in e.children]
  return all([subj not in children_deps for subj in ["csubj"]])

def _get_root(doc: Doc):
  roots = [r for r in doc if r.dep_=="ROOT"]
  assert(len(roots)>=1)
  root = roots[0]
  return root

def get_tree(doc: Doc, nlp: Language):
  root = _get_root(doc)
  
  if _is_noun_like(root):
    xs = ["nsubj", "cop", "obl", "punct", "dep", "aux"]
    children = [r for r in root.children if r.dep_ in xs]
    store = form_children_info(doc, children)

    copula_tree = get_subtree_except_xs(root, xs)
    copula_ranges = [e.i for e in copula_tree]
    start, end = min(copula_ranges), max(copula_ranges)+1
    sub = doc[start:end]
    _info = {
      "start": start,
      "end": end,
      "start_char": sub.start_char,
      "end_char": sub.end_char,
      "text": sub.text,
      "element": root.i, 
      "is_root": True,
    }
    store.append(_info)
  elif _is_verb_like(root):
    children = [r for r in root.children]
    store = form_children_info(doc, children)

    start, end = root.i, root.i+1
    sub = doc[start:end]
    _info = {
      "start": start,
      "end": end,
      "start_char": sub.start_char,
      "end_char": sub.end_char,
      "text": sub.text,
      "element": root.i, 
      "is_root": True,
    }
    store.append(_info)
  else:
    store = []

  ordered_store = sorted(store, key=lambda d: d["start"]) 
  return ordered_store 




from spacy.language import Language 
from spacy.tokens import Doc
from .morph_translation import trans_morph

_es2c = {
  "cop": "系动词",
  "obj": "宾语",
  "iobj": "间接宾语",
  "nsubj": "主语",
  "aux": "助动词",
  "advmod": "状语",
  "advcl": "状语",
  "obl": "状语",
  "expl:pv": "自复代词",
  "conj": "并列句",
}


def _get_root_explanation(doc, t, structure):
  ele = doc[t["element"]]
  deps = [s.dep_ for s in ele.children]
 
  if "cop" in deps:
    return "表语"
  if ele.pos_ in ["AUX", "VERB"]:
    return "动词:" + trans_morph(ele.morph)
  else:
    return ele.dep_

def _analyze(doc, t, structure):
  dep = doc[t["element"]].dep_
  explanation = None

  if dep == "ROOT":
    explanation = _get_root_explanation(doc, t, structure)  
  elif dep in _es2c.keys():
    explanation = _es2c[dep]
  elif dep in ["punct", "dep"]:
    explanation = None
  else:
    explanation = dep
  t["explanation"] = explanation
  return t

def es_trs(doc, structure):
  structure_with_explanation = []
  for t in structure:
    _t = _analyze(doc, t, structure)
    structure_with_explanation.append(_t)
  return structure_with_explanation



from spacy.language import Language 
from spacy.tokens import Doc

_d2c = {
  "nsubj": "主语",
  "csubj": "主语",
  "nsubjpass": "主语", 
  "advmod": "状语",
  "dative": "间接宾语",
  "pobj": "宾语",
  "intj": "感叹语",
  "aux": "助动词",
  "neg": "否定词",
  "auxpass": "助动词",
  "cc": "连词",
  "preconj": "连词",
  "acomp": "表语",
  "expl": "形式主语",
  "agent": "状语",
  "mark": "连词",
  "prt": "小品词",
  "dep": "其他",
}

verb_tags = {
  "VBZ": "动词第三人称单数",
  "VBN": "动词过去分词",
  "VBG": "动词现在分词",
  "VBD": "动词过去式",
  "VB": "动词原形",
  "VBP": "动词原形",
}

cop_verbs = ["be", "prove", "seem", "remain", "get", "look", "feel", "sound", "turn", "smell", "taste", "become", "appear", "grow", "go", "run", "fall", "stay", "keep", "hold", "rest", "emerge"]

svp_enumeration = ["put forth", "come down", "put down", "put aside", "fall through"]

def _get_root_explanation(doc, r, structure):
  ele = doc[r["element"]]
  #deps = [doc[s["element"]].dep_ for s in structure]
  deps = [t.dep_ for t in ele.children]
  explanation = None
  if ele.pos_ not in ["VERB", "AUX"]:
    explanation = None
  elif ele.lemma_=="have" and doc[r["end"]-1].text=="to":
    explanation = "半助动词"
  elif r["end"]-r["start"] > 1:
    explanation = "动词成语"
  elif "expl" in deps and ele.lemma_ == "be": # There be
    explanation = verb_tags[ele.tag_]
  elif ele.text == "going" and "aux" in deps and doc[ele.i+1].text == "to":
    explanation = verb_tags[ele.tag_]
  elif ele.lemma_ in cop_verbs and any([e in ["attr", "oprd", "acomp", "xcomp", "prep"] for e in deps]):  
    explanation = "系动词"
  elif ele.pos_ == "AUX" and any([e in ["advmod", "ccomp"] for e in deps]):  
    explanation = "系动词"
  elif ele.tag_ in verb_tags.keys():
    explanation = verb_tags[ele.tag_]
  else:
    explanation = ele.tag_ 
  return explanation

def _get_aux_explanation(doc, c, structure):
  ele = doc[c["element"]]
  if ele.text == "to":
    explanation = "介词"
  else:
    explanation = "助动词"
  return explanation
  
def _get_conj_explanation(doc, c, structure):
  explanation = None
  ele = doc[c["element"]]
  if c["end"] - c["start"] == 1 and doc[c["element"]].pos_ in ["VERB", "AUX"] and ele.tag_ in verb_tags.keys():
    explanation = verb_tags[ele.tag_]
  else:
    explanation = "并列连句"
  return explanation

def _get_oprd_explanation(doc, o, structure):
  deps = [doc[s["element"]].dep_ for s in structure]
  explanation = None
  if any([e in ["dobj", "auxpass"] for e in deps]):  
    explanation = "补语"
  else:
    explanation = "表语"
  return explanation

def _get_attr_explanation(doc, o, structure):
  deps = [doc[s["element"]].dep_ for s in structure]
  explanation = None
  if any([e in ["expl"] for e in deps]):  
    explanation = "主语"
  else:
    explanation = "表语"
  return explanation

def _get_npadvmod_explanation(doc, t, structure):
  explanation = None
  if t["end"] - t["start"] == 1 and doc[t["element"]].pos_ in ["PROPN", "NOUN"]:
    explanation = "其他"
  else:
    explanation = "状语"
  return explanation
  
def _get_prep_explanation(doc, t, structure):
  root = doc[t["element"]].head
  explanation = None
  deps = [s.dep_ for s in root.children]
  explanation = None
  if root.lemma_ in cop_verbs and t["end"]-t["start"]==1 and all([e not in ["attr", "oprd", "acomp"] for e in deps]):
    explanation = "表语"
  elif t["end"]-t["start"]==1:
    explanation = "介词"
  else:
    explanation = "状语"
  return explanation

def _get_dobj_explanation(doc, t, structure):
  deps = [doc[s["element"]].dep_ for s in structure]
  explanation = None
  if any([e in ["dative"] for e in deps]):  
    explanation = "直接宾语"
  elif any([e in ["ccomp"] for e in deps]):  
    #explanation = "间接宾语"
    explanation = "宾语" 
  else:
    explanation = "宾语"
  return explanation
 
def _get_xcomp_explanation(doc, t, structure):
  root = doc[t["element"]].head
  deps = [doc[s["element"]].dep_ for s in structure]
  explanation = None
  if root.lemma_ in cop_verbs and t["end"]-t["start"]==1:# and all([e not in ["attr", "oprd", "acomp", "prep"] for e in deps]):
    explanation = "表语"
  elif "dobj" in deps and doc[t["element"]].tag_ in ["VBG"]:
    explanation = "补语"
  elif "dobj" in deps:
    explanation = "直接宾语"
  else:
    explanation = "其他"
  return explanation
 
def _get_ccomp_explanation(doc, t, structure): # TBD
  #deps = [doc[s["element"]].dep_ for s in structure]
  ele  = doc[t["element"]]
  ele_children_deps = [c.dep_ for c in ele.children]
  root = doc[t["element"]].head
  deps = [doc[s["element"]].dep_ for s in structure]
  explanation = None
  if "nsubj" in ele_children_deps and t["element"] < root.i:
    explanation = "并列句"
  elif root.pos_ == "AUX":
    explanation = "表语"
  elif "dobj" in deps:
    explanation = "直接宾语"
  else:
    explanation = "其他"
  return explanation
 
def _get_advcl_explanation(doc, t, structure):
  deps = [doc[s["element"]].dep_ for s in structure]
  root = doc[t["element"]].head
  #cop_deps = ["acomp", "prep", "attr", "oprd"]
  #if root.pos_ == "AUX" and all([d not in deps for d in cop_deps])
  #  explanation = "表语"
  #else:
  explanation = "状语"
  return explanation
 
def _get_advmod_explanation(doc, t, structure):
  deps = [doc[s["element"]].dep_ for s in structure]
  root = doc[t["element"]].head
  cop_deps = ["acomp", "prep", "attr", "oprd"]
  if root.pos_ == "AUX" and all([d not in deps for d in cop_deps]) and doc[t["element"]].pos_ not in ["SCONJ"]:
    explanation = "表语"
  elif t["end"] - t["start"] == 1 and " ".join([root.lemma_, t["text"]]) in svp_enumeration:
    explanation = "小品词"
  else:
    explanation = "状语"
  return explanation
 
def _analyze(doc, t, structure):
  #dep = doc[t["element"]].dep_
  dep = t["semantic_dep"]
  name = "_get_{}_explanation".format(dep.lower()) 
  if dep in ["punct"]:
    explanation = None
  elif name in globals().keys():
    explanation = globals()[name](doc, t, structure)
  elif dep in _d2c.keys():
    explanation = _d2c[dep]
  else:
    explanation = dep
  t["explanation"] = explanation
  return t

def en_trs(doc, structure):
  structure_with_explanation = []
  for t in structure:
    _t = _analyze(doc, t, structure)
    structure_with_explanation.append(_t)
  return structure_with_explanation
